# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for fuzzer_stats."""

import datetime
import json
import os
import re
import unittest

import mock
import six

from clusterfuzz._internal.bot.tasks import fuzz_task
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.metrics import fuzzer_stats
from clusterfuzz._internal.tests.test_libs import helpers
from clusterfuzz._internal.tests.test_libs import test_utils


def sanitize_sql(s):
  """Sanitize the sql by removing all new lines and surrounding whitespace."""
  s = re.sub('[ \\s\n\r]*\n[ \\s\n\r]*', ' ', s, flags=re.MULTILINE)
  s = re.sub('\\([ \t]+', '(', s)
  s = re.sub('[ \t]+\\)', ')', s)
  return s.strip()


@test_utils.with_cloud_emulators('datastore')
class FuzzerStatsTest(unittest.TestCase):
  """Fuzzer stats tests."""

  def setUp(self):
    helpers.patch_environ(self)
    data_types.Fuzzer(name='parent').put()

    data_types.FuzzTarget(engine='parent', binary='child').put()
    data_types.FuzzTargetJob(
        engine='parent',
        fuzz_target_name='parent_child',
        job='test_job',
        last_run=datetime.datetime.utcnow()).put()

    helpers.patch(self, [
        'clusterfuzz._internal.google_cloud_utils.storage.write_data',
    ])

  def test_upload_testcase_run(self):
    """Tests uploading of TestcaseRun."""
    testcase_run_0 = fuzzer_stats.TestcaseRun('fuzzer', 'job', 123,
                                              1472846341.017923)
    testcase_run_1 = fuzzer_stats.TestcaseRun('fuzzer', 'job', 123,
                                              1472846341.017923)

    testcase_run_0['stat'] = 1000
    testcase_run_1['stat'] = 2000

    fuzzer_stats.upload_stats(
        [testcase_run_0, testcase_run_1], filename='upload.json')

    self.mock.write_data.assert_called_once_with(
        b'{"fuzzer": "fuzzer", "job": "job", "build_revision": 123, '
        b'"timestamp": 1472846341.017923, "kind": "TestcaseRun", '
        b'"stat": 1000}\n'
        b'{"fuzzer": "fuzzer", "job": "job", "build_revision": 123, '
        b'"timestamp": 1472846341.017923, "kind": "TestcaseRun", '
        b'"stat": 2000}',
        'gs://test-bigquery-bucket/fuzzer/TestcaseRun/date/20160902/upload.json'
    )

  def tests_upload_testcase_run_with_source(self):
    """Test uploading testcase run with source."""
    os.environ['STATS_SOURCE'] = 'custom_source'
    testcase_run = fuzzer_stats.TestcaseRun('fuzzer', 'job', 123,
                                            1472846341.017923)
    fuzzer_stats.upload_stats([testcase_run], filename='upload.json')
    self.mock.write_data.assert_called_once_with(
        b'{"fuzzer": "fuzzer", "job": "job", "build_revision": 123, '
        b'"timestamp": 1472846341.017923, "kind": "TestcaseRun", '
        b'"source": "custom_source"}',
        'gs://test-bigquery-bucket/fuzzer/TestcaseRun/date/20160902/upload.json'
    )

  def test_upload_testcase_run_child(self):
    """Tests uploading of Testcaserun for a child fuzzer."""
    testcase_run_0 = fuzzer_stats.TestcaseRun('parent_child', 'job', 123,
                                              1472846341.017923)
    testcase_run_0['stat'] = 1000

    fuzzer_stats.upload_stats([testcase_run_0], filename='upload.json')
    self.mock.write_data.assert_called_once_with(
        b'{"fuzzer": "parent_child", "job": "job", "build_revision": 123, '
        b'"timestamp": 1472846341.017923, "kind": "TestcaseRun", '
        b'"stat": 1000}',
        'gs://test-bigquery-bucket/parent/TestcaseRun/date/20160902/upload.json'
    )

  def test_upload_testcase_run_2_days(self):
    """Tests uploading TestcaseRuns that span multiple days."""
    testcase_run_0 = fuzzer_stats.TestcaseRun('fuzzer', 'job', 123,
                                              1472846341.017923)
    testcase_run_1 = fuzzer_stats.TestcaseRun('fuzzer', 'job', 123,
                                              1472846345.017923)
    testcase_run_2 = fuzzer_stats.TestcaseRun('fuzzer', 'job', 123,
                                              1472932741.017923)
    testcase_run_3 = fuzzer_stats.TestcaseRun('fuzzer', 'job', 123,
                                              1472932745.017923)
    testcase_run_0['stat'] = 1000
    testcase_run_1['stat'] = 2000
    testcase_run_2['stat'] = 3000
    testcase_run_3['stat'] = 4000

    fuzzer_stats.upload_stats(
        [testcase_run_0, testcase_run_1, testcase_run_2, testcase_run_3],
        filename='upload.json')

    expected_calls = [
        mock.call(
            b'{"fuzzer": "fuzzer", "job": "job", "build_revision": 123, '
            b'"timestamp": 1472846341.017923, "kind": "TestcaseRun", '
            b'"stat": 1000}\n'
            b'{"fuzzer": "fuzzer", "job": "job", "build_revision": 123, '
            b'"timestamp": 1472846345.017923, "kind": "TestcaseRun", '
            b'"stat": 2000}',
            'gs://test-bigquery-bucket/fuzzer/TestcaseRun/date/20160902/'
            'upload.json'),
        mock.call(
            b'{"fuzzer": "fuzzer", "job": "job", "build_revision": 123, '
            b'"timestamp": 1472932741.017923, "kind": "TestcaseRun", '
            b'"stat": 3000}\n'
            b'{"fuzzer": "fuzzer", "job": "job", "build_revision": 123, '
            b'"timestamp": 1472932745.017923, "kind": "TestcaseRun", '
            b'"stat": 4000}',
            'gs://test-bigquery-bucket/fuzzer/TestcaseRun/date/20160903/'
            'upload.json'),
    ]

    six.assertCountEqual(self, self.mock.write_data.call_args_list,
                         expected_calls)

  def test_upload_job_run(self):
    """Tests uploading of JobRun."""
    crashes = [{
        'is_new': False,
        'count': 2,
        'crash_type': 't1',
        'crash_state': 's1',
        'security_flag': True
    }]
    fuzzer_run = fuzzer_stats.JobRun('fuzzer', 'job', 123, 1472846341.017923,
                                     9001, 0, 1, crashes)

    fuzzer_stats.upload_stats([fuzzer_run], filename='upload.json')

    self.assertEqual(1, self.mock.write_data.call_count)
    self.assertEqual({
        'kind': 'JobRun',
        'known_crashes': 1,
        'timestamp': 1472846341.017923,
        'job': 'job',
        'fuzzer': 'fuzzer',
        'new_crashes': 0,
        'build_revision': 123,
        'testcases_executed': 9001,
        'crashes': crashes
    }, json.loads(self.mock.write_data.call_args[0][0]))

    self.assertEqual(
        'gs://test-bigquery-bucket/fuzzer/JobRun/date/20160902/upload.json',
        self.mock.write_data.call_args[0][1])

  @mock.patch('os.path.exists')
  def test_testcase_run_read_from_disk(self, mock_path_exists):
    """Tests TestcaseRun deserialization."""
    read_data = ('{"stat": 1000, "timestamp": 1472846341.017923, '
                 '"kind": "TestcaseRun", "job": "job", "fuzzer": "fuzzer", '
                 '"build_revision": 123}')

    mock_path_exists.return_value = True
    m = mock.mock_open(read_data=read_data)
    with mock.patch('clusterfuzz._internal.metrics.fuzzer_stats.open', m):
      testcase_run = fuzzer_stats.TestcaseRun.read_from_disk('fake_path')

    self.assertIsNotNone(testcase_run)
    self.assertEqual(testcase_run.kind, 'TestcaseRun')
    self.assertEqual(testcase_run.fuzzer, 'fuzzer')
    self.assertEqual(testcase_run.job, 'job')
    self.assertEqual(testcase_run.build_revision, 123)
    self.assertEqual(testcase_run.timestamp, 1472846341.017923)
    self.assertEqual(testcase_run['stat'], 1000)

  def test_testcase_run_write_to_disk(self):
    """Tests TestcaseRun serialization."""
    testcase_run = fuzzer_stats.TestcaseRun('fuzzer', 'job', 123,
                                            1472846341.017923)

    m = mock.mock_open()
    with mock.patch('clusterfuzz._internal.metrics.fuzzer_stats.open', m):
      fuzzer_stats.TestcaseRun.write_to_disk(testcase_run, 'fake_path')

    handle = m()
    handle.write.assert_called_once_with(
        '{"fuzzer": "fuzzer", "job": "job", "build_revision": 123, '
        '"timestamp": 1472846341.017923, "kind": "TestcaseRun"}')

  def test_job_run_from_json(self):
    """Tests JobRun deserialization."""
    data = json.dumps({
        'kind': 'JobRun',
        'known_crashes': 1,
        'timestamp': 1472846341.017923,
        'job': 'job',
        'fuzzer': 'fuzzer',
        'new_crashes': 0,
        'build_revision': 123,
        'testcases_executed': 9001,
        'crashes': [{
            'test': 'crash'
        }]
    })
    job_run = fuzzer_stats.BaseRun.from_json(data)
    self.assertIsNotNone(job_run)
    self.assertEqual(job_run.kind, 'JobRun')
    self.assertEqual(job_run.fuzzer, 'fuzzer')
    self.assertEqual(job_run.job, 'job')
    self.assertEqual(job_run.build_revision, 123)
    self.assertEqual(job_run.timestamp, 1472846341.017923)
    self.assertEqual(job_run['new_crashes'], 0)
    self.assertEqual(job_run['known_crashes'], 1)
    self.assertEqual(job_run['testcases_executed'], 9001)
    self.assertEqual(job_run['crashes'], [{'test': 'crash'}])

  @mock.patch(
      'clusterfuzz._internal.metrics.fuzzer_stats.TestcaseRun.read_from_disk')
  def test_fuzz_task_upload_testcase_run_stats_builtin_fuzzer(
      self, mock_read_from_disk_new):
    """Tests that fuzz_task.read_and_upload_testcase_run_stats uploads stats."""
    testcase_run = fuzzer_stats.TestcaseRun('placeholder', 'placeholder', 0,
                                            1472846341.017923)
    testcase_run['stat'] = 9001

    mock_read_from_disk_new.return_value = testcase_run
    fuzz_task.read_and_upload_testcase_run_stats(
        'libFuzzer', 'libFuzzer_fuzzer1', 'job', 123, ['fake_path'])

    self.assertEqual(1, self.mock.write_data.call_count)

    self.assertEqual(
        b'{"fuzzer": "libFuzzer_fuzzer1", "job": "job", "build_revision": 123, '
        b'"timestamp": 1472846341.017923, "kind": "TestcaseRun", "stat": 9001}',
        self.mock.write_data.call_args[0][0])

  @mock.patch(
      'clusterfuzz._internal.metrics.fuzzer_stats.TestcaseRun.read_from_disk')
  def test_fuzz_task_upload_testcase_run_stats_blackbox_fuzzer(
      self, mock_read_from_disk_new):
    """Tests that fuzz_task.read_and_upload_testcase_run_stats uploads stats."""
    testcase_run = fuzzer_stats.TestcaseRun('placeholder', 'placeholder', 0,
                                            1472846341.017923)
    testcase_run['stat'] = 9001

    mock_read_from_disk_new.return_value = testcase_run
    fuzz_task.read_and_upload_testcase_run_stats(
        'blackbox_fuzzer', 'blackbox_fuzzer', 'job', 123, ['fake_path'])

    self.assertEqual(0, self.mock.write_data.call_count)

  def test_fuzz_task_upload_job_run_stats(self):
    """Tests that fuzz_task.upload_job_run_stats uploads stats."""
    groups = [
        mock.Mock(
            crashes=[mock.Mock(), mock.Mock()],
            main_crash=mock.Mock(
                crash_type='t1', crash_state='s1', security_flag=True)),
        mock.Mock(
            crashes=[mock.Mock()],
            main_crash=mock.Mock(
                crash_type='t2', crash_state='s2', security_flag=False)),
    ]
    groups[0].is_new.return_value = False
    groups[1].is_new.return_value = True

    fuzz_task.upload_job_run_stats('fuzzer', 'job', 123, 1472846341.017923, 1,
                                   2, 1337, groups)
    self.assertEqual(1, self.mock.write_data.call_count)
    self.assertEqual({
        'kind':
            'JobRun',
        'known_crashes':
            2,
        'timestamp':
            1472846341.017923,
        'job':
            'job',
        'fuzzer':
            'fuzzer',
        'new_crashes':
            1,
        'build_revision':
            123,
        'testcases_executed':
            1337,
        'crashes': [
            {
                'is_new': False,
                'count': 2,
                'crash_type': 't1',
                'crash_state': 's1',
                'security_flag': True
            },
            {
                'is_new': True,
                'count': 1,
                'crash_type': 't2',
                'crash_state': 's2',
                'security_flag': False
            },
        ]
    }, json.loads(self.mock.write_data.call_args[0][0]))


@test_utils.with_cloud_emulators('datastore')
class BigQueryStatsTests(unittest.TestCase):
  """BigQuery stats tests."""

  def setUp(self):
    data_types.Fuzzer(name='parent').put()

    data_types.FuzzTarget(engine='parent', binary='child').put()
    data_types.FuzzTargetJob(
        engine='parent',
        fuzz_target_name='parent_child',
        job='test_job',
        last_run=datetime.datetime.utcnow()).put()

  def test_parse_stats_column_fields(self):
    """Tests stats column parsing."""
    fields = fuzzer_stats.parse_stats_column_fields(
        'sum(t.abc), avg(j.abc) as bcd, custom(j.def) as def,  '
        '_EDGE_COV, _FUNC_COV as 123,\n'
        '_COV_REPORT as blahblah, _CORPUS_SIZE as corpus_size, '
        '_CORPUS_BACKUP as corpus_backup')

    self.assertEqual(len(fields), 8)

    self.assertIsInstance(fields[0], fuzzer_stats.QueryField)
    self.assertEqual(fields[0].aggregate_function, 'sum')
    self.assertFalse(fields[0].is_custom())
    self.assertEqual(fields[0].name, 'abc')
    self.assertEqual(fields[0].table_alias, 't')
    # select_alias is defauled to name.
    self.assertEqual(fields[0].select_alias, 'abc')

    self.assertIsInstance(fields[1], fuzzer_stats.QueryField)
    self.assertEqual(fields[1].aggregate_function, 'avg')
    self.assertFalse(fields[1].is_custom())
    self.assertEqual(fields[1].name, 'abc')
    self.assertEqual(fields[1].table_alias, 'j')
    self.assertEqual(fields[1].select_alias, 'bcd')

    self.assertIsInstance(fields[2], fuzzer_stats.QueryField)
    self.assertEqual(fields[2].aggregate_function, 'custom')
    self.assertTrue(fields[2].is_custom())
    self.assertEqual(fields[2].name, 'def')
    self.assertEqual(fields[2].table_alias, 'j')
    self.assertEqual(fields[2].select_alias, 'def')

    self.assertIsInstance(fields[3], fuzzer_stats.BuiltinFieldSpecifier)
    self.assertEqual(fields[3].name, '_EDGE_COV')
    self.assertEqual(fields[3].field_class(), fuzzer_stats.CoverageField)
    self.assertIsNone(fields[3].alias)

    self.assertIsInstance(fields[4], fuzzer_stats.BuiltinFieldSpecifier)
    self.assertEqual(fields[4].name, '_FUNC_COV')
    self.assertEqual(fields[4].field_class(), fuzzer_stats.CoverageField)
    self.assertEqual(fields[4].alias, '123')

    self.assertIsInstance(fields[5], fuzzer_stats.BuiltinFieldSpecifier)
    self.assertEqual(fields[5].name, '_COV_REPORT')
    self.assertEqual(fields[5].field_class(), fuzzer_stats.CoverageReportField)
    self.assertEqual(fields[5].alias, 'blahblah')

    self.assertIsInstance(fields[6], fuzzer_stats.BuiltinFieldSpecifier)
    self.assertEqual(fields[6].name, '_CORPUS_SIZE')
    self.assertEqual(fields[6].field_class(), fuzzer_stats.CorpusSizeField)
    self.assertEqual(fields[6].alias, 'corpus_size')

    self.assertIsInstance(fields[7], fuzzer_stats.BuiltinFieldSpecifier)
    self.assertEqual(fields[7].name, '_CORPUS_BACKUP')
    self.assertEqual(fields[7].field_class(), fuzzer_stats.CorpusBackupField)
    self.assertEqual(fields[7].alias, 'corpus_backup')

    # Test that invalid fields are ignored.
    fields = fuzzer_stats.parse_stats_column_fields(
        'sum(abc)  ,   min(t.bcd)    as bcd   , '
        'sum(t.def) as "1, _EDGE_COV as ""1"')
    self.assertEqual(len(fields), 1)
    self.assertIsInstance(fields[0], fuzzer_stats.QueryField)
    self.assertEqual(fields[0].aggregate_function, 'min')
    self.assertEqual(fields[0].name, 'bcd')
    self.assertEqual(fields[0].table_alias, 't')
    self.assertEqual(fields[0].select_alias, 'bcd')

  def test_query_job_day(self):
    """Tests querying for JobRuns grouped by day."""
    fields = fuzzer_stats.parse_stats_column_fields(
        fuzzer_stats.JobQuery.DEFAULT_FIELDS)
    query = fuzzer_stats.JobQuery('fuzzer_name', ['job_type', 'job_type2'],
                                  fields,
                                  fuzzer_stats.QueryGroupBy.GROUP_BY_DAY,
                                  datetime.date(2016, 10, 1),
                                  datetime.date(2016, 10, 7))

    self.assertEqual(
        sanitize_sql(query.build()),
        sanitize_sql("""
        WITH
          JobRunWithConcatedCrashes AS (
            SELECT
              TIMESTAMP_TRUNC(
                TIMESTAMP_SECONDS(CAST(timestamp AS INT64)), DAY, "UTC"
              ) as date,
              sum(testcases_executed) as testcases_executed,
              ARRAY_CONCAT_AGG(crashes) AS crashes
            FROM
              `test-clusterfuzz`.fuzzer_name_stats.JobRun
            WHERE
              (
                _PARTITIONTIME BETWEEN
                TIMESTAMP_SECONDS(1475280000) AND TIMESTAMP_SECONDS(1475798400)
              ) AND (
                job = \'job_type\' OR job = \'job_type2\'
              )
            GROUP BY date
          ),
          JobRunWithUniqueCrashes AS (
            SELECT
              * EXCEPT(crashes),
              ARRAY(
                SELECT AS STRUCT
                  crash.crash_type,
                  crash.crash_state,
                  crash.security_flag,
                  SUM(count) AS count,
                  MAX(crash.is_new) AS is_new
                FROM
                  UNNEST(crashes) AS crash
                GROUP BY
                  crash.crash_type,
                  crash.crash_state,
                  crash.security_flag
              ) AS crashes
            FROM
              JobRunWithConcatedCrashes
          ),
          JobRunWithSummary AS (
            SELECT
              * EXCEPT(crashes),
              (
                SELECT AS STRUCT
                  IFNULL(SUM(crash.count), 0) AS total,
                  COUNTIF(crash.is_new) AS unique_new,
                  COUNT(crash) AS unique
                FROM
                  UNNEST(crashes) AS crash
              ) AS crash_count
            FROM
              JobRunWithUniqueCrashes
          )

        SELECT
          * EXCEPT(crash_count),
          crash_count.total AS total_crashes,
          crash_count.unique_new AS new_crashes,
          (crash_count.unique - crash_count.unique_new) AS known_crashes
        FROM
          JobRunWithSummary
        """))

  def test_query_job_revision(self):
    """Tests querying for JobRuns grouped by revision."""
    fields = fuzzer_stats.parse_stats_column_fields(
        fuzzer_stats.JobQuery.DEFAULT_FIELDS)
    query = fuzzer_stats.JobQuery('fuzzer_name', ['job_type', 'job_type2'],
                                  fields,
                                  fuzzer_stats.QueryGroupBy.GROUP_BY_REVISION,
                                  datetime.date(2016, 10, 1),
                                  datetime.date(2016, 10, 7))

    self.assertEqual(
        sanitize_sql(query.build()),
        sanitize_sql("""
        WITH
          JobRunWithConcatedCrashes AS (
            SELECT
              build_revision,
              sum(testcases_executed) as testcases_executed,
              ARRAY_CONCAT_AGG(crashes) AS crashes
            FROM
              `test-clusterfuzz`.fuzzer_name_stats.JobRun
            WHERE
              (
                _PARTITIONTIME BETWEEN
                TIMESTAMP_SECONDS(1475280000) AND TIMESTAMP_SECONDS(1475798400)
              ) AND (
                job = \'job_type\' OR job = \'job_type2\'
              )
            GROUP BY build_revision
          ),
          JobRunWithUniqueCrashes AS (
            SELECT
              * EXCEPT(crashes),
              ARRAY(
                SELECT AS STRUCT
                  crash.crash_type,
                  crash.crash_state,
                  crash.security_flag,
                  SUM(count) AS count,
                  MAX(crash.is_new) AS is_new
                FROM
                  UNNEST(crashes) AS crash
                GROUP BY
                  crash.crash_type,
                  crash.crash_state,
                  crash.security_flag
              ) AS crashes
            FROM
              JobRunWithConcatedCrashes
          ),
          JobRunWithSummary AS (
            SELECT
              * EXCEPT(crashes),
              (
                SELECT AS STRUCT
                  IFNULL(SUM(crash.count), 0) AS total,
                  COUNTIF(crash.is_new) AS unique_new,
                  COUNT(crash) AS unique
                FROM
                  UNNEST(crashes) AS crash
              ) AS crash_count
            FROM
              JobRunWithUniqueCrashes
          )

        SELECT
          * EXCEPT(crash_count),
          crash_count.total AS total_crashes,
          crash_count.unique_new AS new_crashes,
          (crash_count.unique - crash_count.unique_new) AS known_crashes
        FROM
          JobRunWithSummary
        """))

  def test_query_job_fuzzer(self):
    """Tests querying for JobRuns grouped by fuzzer."""
    fields = fuzzer_stats.parse_stats_column_fields(
        fuzzer_stats.JobQuery.DEFAULT_FIELDS)
    query = fuzzer_stats.JobQuery('fuzzer_name', ['job_type', 'job_type2'],
                                  fields,
                                  fuzzer_stats.QueryGroupBy.GROUP_BY_FUZZER,
                                  datetime.date(2016, 10, 1),
                                  datetime.date(2016, 10, 7))

    self.assertEqual(
        sanitize_sql(query.build()),
        sanitize_sql("""
        WITH
          JobRunWithConcatedCrashes AS (
            SELECT
              fuzzer,
              sum(testcases_executed) as testcases_executed,
              ARRAY_CONCAT_AGG(crashes) AS crashes
            FROM
              `test-clusterfuzz`.fuzzer_name_stats.JobRun
            WHERE
              (
                _PARTITIONTIME BETWEEN TIMESTAMP_SECONDS(1475280000)
                AND TIMESTAMP_SECONDS(1475798400)
              ) AND (
                job = \'job_type\' OR job = \'job_type2\'
              )
            GROUP BY fuzzer
          ),
          JobRunWithUniqueCrashes AS (
            SELECT
              * EXCEPT(crashes),
              ARRAY(
                SELECT AS STRUCT
                  crash.crash_type,
                  crash.crash_state,
                  crash.security_flag,
                  SUM(count) AS count,
                  MAX(crash.is_new) AS is_new
                FROM
                  UNNEST(crashes) AS crash
                GROUP BY
                  crash.crash_type,
                  crash.crash_state,
                  crash.security_flag
              ) AS crashes
            FROM
              JobRunWithConcatedCrashes
          ),
          JobRunWithSummary AS (
            SELECT
              * EXCEPT(crashes),
              (
                SELECT AS STRUCT
                  IFNULL(SUM(crash.count), 0) AS total,
                  COUNTIF(crash.is_new) AS unique_new,
                  COUNT(crash) AS unique
                FROM
                  UNNEST(crashes) AS crash
              ) AS crash_count
            FROM
              JobRunWithUniqueCrashes
          )

        SELECT
          * EXCEPT(crash_count),
          crash_count.total AS total_crashes,
          crash_count.unique_new AS new_crashes,
          (crash_count.unique - crash_count.unique_new) AS known_crashes
        FROM
          JobRunWithSummary
        """))

  def test_table_query_join(self):
    """Tests basic table query involving a join."""
    stats_columns = """
      sum(j.testcases_executed) as testcases_executed,
      custom(j.total_crashes) as total_crashes,
      custom(j.new_crashes) as new_crashes,
      custom(j.known_crashes) as known_crashes,
      avg(t.average_exec_per_sec) as average_exec_per_sec
    """

    query = fuzzer_stats.TableQuery('fuzzer_name', ['job_type', 'job_type2'],
                                    stats_columns,
                                    fuzzer_stats.QueryGroupBy.GROUP_BY_DAY,
                                    datetime.date(2016, 10, 1),
                                    datetime.date(2016, 10, 7))
    self.assertEqual(
        sanitize_sql(query.build()),
        sanitize_sql("""
        SELECT j.date, * EXCEPT(date) FROM (
          WITH
            JobRunWithConcatedCrashes AS (
              SELECT
                TIMESTAMP_TRUNC(
                  TIMESTAMP_SECONDS(CAST(timestamp AS INT64)), DAY, "UTC"
                ) as date,
                sum(testcases_executed) as testcases_executed,
                ARRAY_CONCAT_AGG(crashes) AS crashes
              FROM
                `test-clusterfuzz`.fuzzer_name_stats.JobRun
              WHERE
                (
                  _PARTITIONTIME BETWEEN TIMESTAMP_SECONDS(1475280000)
                  AND TIMESTAMP_SECONDS(1475798400)
                ) AND (
                  job = \'job_type\' OR job = \'job_type2\'
                )
              GROUP BY date
            ),
            JobRunWithUniqueCrashes AS (
              SELECT
                * EXCEPT(crashes),
                ARRAY(
                  SELECT AS STRUCT
                    crash.crash_type,
                    crash.crash_state,
                    crash.security_flag,
                    SUM(count) AS count,
                    MAX(crash.is_new) AS is_new
                  FROM
                    UNNEST(crashes) AS crash
                  GROUP BY
                    crash.crash_type,
                    crash.crash_state,
                    crash.security_flag
                ) AS crashes
              FROM
                JobRunWithConcatedCrashes
            ),
            JobRunWithSummary AS (
              SELECT
                * EXCEPT(crashes),
                (
                  SELECT AS STRUCT
                    IFNULL(SUM(crash.count), 0) AS total,
                    COUNTIF(crash.is_new) AS unique_new,
                    COUNT(crash) AS unique
                  FROM
                    UNNEST(crashes) AS crash
                ) AS crash_count
              FROM
                JobRunWithUniqueCrashes
            )

          SELECT
            * EXCEPT(crash_count),
            crash_count.total AS total_crashes,
            crash_count.unique_new AS new_crashes,
            (crash_count.unique - crash_count.unique_new) AS known_crashes
          FROM
            JobRunWithSummary
        ) as j INNER JOIN (
          SELECT
            TIMESTAMP_TRUNC(
              TIMESTAMP_SECONDS(CAST(timestamp AS INT64)), DAY, "UTC"
            ) as date,
            avg(average_exec_per_sec) as average_exec_per_sec
          FROM `test-clusterfuzz`.fuzzer_name_stats.TestcaseRun
          WHERE
            (
              _PARTITIONTIME BETWEEN TIMESTAMP_SECONDS(1475280000) AND
              TIMESTAMP_SECONDS(1475798400)
            ) AND (
              job = 'job_type' OR job = 'job_type2'
            )
          GROUP BY date
        ) as t ON j.date = t.date
        """))

  def test_table_query_single(self):
    """Tests basic table query involving single subquery."""
    stats_columns = """
      sum(j.testcases_executed) as testcases_executed,
      custom(j.total_crashes) as total_crashes,
      custom(j.new_crashes) as new_crashes,
      custom(j.known_crashes) as known_crashes
    """

    query = fuzzer_stats.TableQuery('fuzzer_name', ['job_type'], stats_columns,
                                    fuzzer_stats.QueryGroupBy.GROUP_BY_DAY,
                                    datetime.date(2016, 10, 1),
                                    datetime.date(2016, 10, 7))

    self.assertEqual(
        sanitize_sql(query.build()),
        sanitize_sql("""
        SELECT j.date, * EXCEPT(date) FROM (
          WITH
            JobRunWithConcatedCrashes AS (
              SELECT
                TIMESTAMP_TRUNC(
                  TIMESTAMP_SECONDS(CAST(timestamp AS INT64)), DAY, "UTC"
                ) as date,
                sum(testcases_executed) as testcases_executed,
                ARRAY_CONCAT_AGG(crashes) AS crashes
              FROM
                `test-clusterfuzz`.fuzzer_name_stats.JobRun
              WHERE
                (
                  _PARTITIONTIME BETWEEN TIMESTAMP_SECONDS(1475280000) AND
                  TIMESTAMP_SECONDS(1475798400)
                ) AND (
                  job = \'job_type\'
                )
              GROUP BY date
            ),
            JobRunWithUniqueCrashes AS (
              SELECT
                * EXCEPT(crashes),
                ARRAY(
                  SELECT AS STRUCT
                    crash.crash_type,
                    crash.crash_state,
                    crash.security_flag,
                    SUM(count) AS count,
                    MAX(crash.is_new) AS is_new
                  FROM
                    UNNEST(crashes) AS crash
                  GROUP BY
                    crash.crash_type,
                    crash.crash_state,
                    crash.security_flag
                ) AS crashes
              FROM
                JobRunWithConcatedCrashes
            ),
            JobRunWithSummary AS (
              SELECT
                * EXCEPT(crashes),
                (
                  SELECT AS STRUCT
                    IFNULL(SUM(crash.count), 0) AS total,
                    COUNTIF(crash.is_new) AS unique_new,
                    COUNT(crash) AS unique
                  FROM
                    UNNEST(crashes) AS crash
                ) AS crash_count
              FROM
                JobRunWithUniqueCrashes
            )

          SELECT
            * EXCEPT(crash_count),
            crash_count.total AS total_crashes,
            crash_count.unique_new AS new_crashes,
            (crash_count.unique - crash_count.unique_new) AS known_crashes
          FROM
            JobRunWithSummary
        ) as j
        """))

  def test_table_query_group_fuzzer(self):
    """Tests table query grouping by fuzzer."""
    stats_columns = """
      sum(j.testcases_executed) as testcases_executed,
      custom(j.total_crashes) as total_crashes,
      custom(j.new_crashes) as new_crashes,
      custom(j.known_crashes) as known_crashes,
      avg(t.average_exec_per_sec) as average_exec_per_sec
    """

    query = fuzzer_stats.TableQuery('parent_child', ['test_job', 'test_job2'],
                                    stats_columns,
                                    fuzzer_stats.QueryGroupBy.GROUP_BY_FUZZER,
                                    datetime.date(2016, 10, 1),
                                    datetime.date(2016, 10, 7))
    self.assertEqual(
        sanitize_sql(query.build()),
        sanitize_sql("""
        SELECT j.fuzzer, * EXCEPT(fuzzer) FROM (
          WITH
            JobRunWithConcatedCrashes AS (
              SELECT
                fuzzer,
                sum(testcases_executed) as testcases_executed,
                ARRAY_CONCAT_AGG(crashes) AS crashes
              FROM
                `test-clusterfuzz`.parent_stats.JobRun
              WHERE
                (
                  _PARTITIONTIME BETWEEN TIMESTAMP_SECONDS(1475280000) AND
                  TIMESTAMP_SECONDS(1475798400)
                ) AND (
                  job = \'test_job\' OR job = \'test_job2\'
                )
                AND fuzzer = \'parent_child\'
              GROUP BY fuzzer
            ),
            JobRunWithUniqueCrashes AS (
              SELECT
                * EXCEPT(crashes),
                ARRAY(
                  SELECT AS STRUCT
                    crash.crash_type,
                    crash.crash_state,
                    crash.security_flag,
                    SUM(count) AS count,
                    MAX(crash.is_new) AS is_new
                  FROM
                    UNNEST(crashes) AS crash
                  GROUP BY
                    crash.crash_type,
                    crash.crash_state,
                    crash.security_flag
                ) AS crashes
              FROM
                JobRunWithConcatedCrashes
            ),
            JobRunWithSummary AS (
              SELECT
                * EXCEPT(crashes),
                (
                  SELECT AS STRUCT
                    IFNULL(SUM(crash.count), 0) AS total,
                    COUNTIF(crash.is_new) AS unique_new,
                    COUNT(crash) AS unique
                  FROM
                    UNNEST(crashes) AS crash
                ) AS crash_count
              FROM
                JobRunWithUniqueCrashes
            )

          SELECT
            * EXCEPT(crash_count),
            crash_count.total AS total_crashes,
            crash_count.unique_new AS new_crashes,
            (crash_count.unique - crash_count.unique_new) AS known_crashes
          FROM
            JobRunWithSummary
        ) as j INNER JOIN (
          SELECT
            fuzzer,
            avg(average_exec_per_sec) as average_exec_per_sec
          FROM `test-clusterfuzz`.parent_stats.TestcaseRun
          WHERE
            (
              _PARTITIONTIME BETWEEN TIMESTAMP_SECONDS(1475280000) AND
              TIMESTAMP_SECONDS(1475798400)
            ) AND (
              job = \'test_job\' OR job = \'test_job2\'
            ) AND fuzzer = \'parent_child\'
          GROUP BY fuzzer
        ) as t ON j.fuzzer = t.fuzzer
        """))

    # Don't specify a job.
    query = fuzzer_stats.TableQuery('parent_child', None, stats_columns,
                                    fuzzer_stats.QueryGroupBy.GROUP_BY_FUZZER,
                                    datetime.date(2016, 10, 1),
                                    datetime.date(2016, 10, 7))
    self.assertEqual(
        sanitize_sql(query.build()),
        sanitize_sql("""
        SELECT j.fuzzer, * EXCEPT(fuzzer) FROM (
          WITH
            JobRunWithConcatedCrashes AS (
              SELECT
                fuzzer,
                sum(testcases_executed) as testcases_executed,
                ARRAY_CONCAT_AGG(crashes) AS crashes
              FROM
                `test-clusterfuzz`.parent_stats.JobRun
              WHERE
                (
                  _PARTITIONTIME BETWEEN TIMESTAMP_SECONDS(1475280000) AND
                  TIMESTAMP_SECONDS(1475798400)
                )
                AND fuzzer = \'parent_child\'
              GROUP BY fuzzer
            ),
            JobRunWithUniqueCrashes AS (
              SELECT
                * EXCEPT(crashes),
                ARRAY(
                  SELECT AS STRUCT
                    crash.crash_type,
                    crash.crash_state,
                    crash.security_flag,
                    SUM(count) AS count,
                    MAX(crash.is_new) AS is_new
                  FROM
                    UNNEST(crashes) AS crash
                  GROUP BY
                    crash.crash_type,
                    crash.crash_state,
                    crash.security_flag
                ) AS crashes
              FROM
                JobRunWithConcatedCrashes
            ),
            JobRunWithSummary AS (
              SELECT
                * EXCEPT(crashes),
                (
                  SELECT AS STRUCT
                    IFNULL(SUM(crash.count), 0) AS total,
                    COUNTIF(crash.is_new) AS unique_new,
                    COUNT(crash) AS unique
                  FROM
                    UNNEST(crashes) AS crash
                ) AS crash_count
              FROM
                JobRunWithUniqueCrashes
            )

          SELECT
            * EXCEPT(crash_count),
            crash_count.total AS total_crashes,
            crash_count.unique_new AS new_crashes,
            (crash_count.unique - crash_count.unique_new) AS known_crashes
          FROM
            JobRunWithSummary
        ) as j INNER JOIN (
          SELECT
            fuzzer,
            avg(average_exec_per_sec) as average_exec_per_sec
          FROM `test-clusterfuzz`.parent_stats.TestcaseRun
          WHERE
            (
              _PARTITIONTIME BETWEEN TIMESTAMP_SECONDS(1475280000) AND
              TIMESTAMP_SECONDS(1475798400)
            )
            AND fuzzer = \'parent_child\'
          GROUP BY fuzzer
        ) as t ON j.fuzzer = t.fuzzer
        """))

  def test_table_query_group_job(self):
    """Tests grouping by job."""
    stats_columns = """
      sum(j.testcases_executed) as testcases_executed,
      custom(j.total_crashes) as total_crashes,
      custom(j.new_crashes) as new_crashes,
      custom(j.known_crashes) as known_crashes,
      avg(t.average_exec_per_sec) as average_exec_per_sec
    """

    query = fuzzer_stats.TableQuery('parent_child', None, stats_columns,
                                    fuzzer_stats.QueryGroupBy.GROUP_BY_JOB,
                                    datetime.date(2016, 10, 1),
                                    datetime.date(2016, 10, 7))
    self.assertEqual(
        sanitize_sql(query.build()),
        sanitize_sql("""
        SELECT j.job, * EXCEPT(job) FROM (
          WITH
            JobRunWithConcatedCrashes AS (
              SELECT
                job,
                sum(testcases_executed) as testcases_executed,
                ARRAY_CONCAT_AGG(crashes) AS crashes
              FROM
                `test-clusterfuzz`.parent_stats.JobRun
              WHERE
                (
                  _PARTITIONTIME BETWEEN TIMESTAMP_SECONDS(1475280000) AND
                  TIMESTAMP_SECONDS(1475798400)
                )
                AND fuzzer = \'parent_child\'
              GROUP BY job
            ),
            JobRunWithUniqueCrashes AS (
              SELECT
                * EXCEPT(crashes),
                ARRAY(
                  SELECT AS STRUCT
                    crash.crash_type,
                    crash.crash_state,
                    crash.security_flag,
                    SUM(count) AS count,
                    MAX(crash.is_new) AS is_new
                  FROM
                    UNNEST(crashes) AS crash
                  GROUP BY
                    crash.crash_type,
                    crash.crash_state,
                    crash.security_flag
                ) AS crashes
              FROM
                JobRunWithConcatedCrashes
            ),
            JobRunWithSummary AS (
              SELECT
                * EXCEPT(crashes),
                (
                  SELECT AS STRUCT
                    IFNULL(SUM(crash.count), 0) AS total,
                    COUNTIF(crash.is_new) AS unique_new,
                    COUNT(crash) AS unique
                  FROM
                    UNNEST(crashes) AS crash
                ) AS crash_count
              FROM
                JobRunWithUniqueCrashes
            )

          SELECT
            * EXCEPT(crash_count),
            crash_count.total AS total_crashes,
            crash_count.unique_new AS new_crashes,
            (crash_count.unique - crash_count.unique_new) AS known_crashes
          FROM
            JobRunWithSummary
        ) as j INNER JOIN (
          SELECT
            job,
            avg(average_exec_per_sec) as average_exec_per_sec
          FROM `test-clusterfuzz`.parent_stats.TestcaseRun
          WHERE
            (
              _PARTITIONTIME BETWEEN TIMESTAMP_SECONDS(1475280000) AND
              TIMESTAMP_SECONDS(1475798400)
            ) AND
            fuzzer = \'parent_child\'
          GROUP BY job
        ) as t ON j.job = t.job
        """))

  def test_table_query_group_time(self):
    """Tests table query grouping by fuzzer."""
    stats_columns = """
      sum(j.testcases_executed) as testcases_executed,
      custom(j.total_crashes) as total_crashes,
      custom(j.new_crashes) as new_crashes,
      custom(j.known_crashes) as known_crashes,
      avg(t.average_exec_per_sec) as average_exec_per_sec
    """

    query = fuzzer_stats.TableQuery('parent_child', ['test_job', 'test_job2'],
                                    stats_columns,
                                    fuzzer_stats.QueryGroupBy.GROUP_BY_TIME,
                                    datetime.date(2016, 10, 1),
                                    datetime.date(2016, 10, 7))
    self.assertEqual(
        sanitize_sql(query.build()),
        sanitize_sql("""
        SELECT
          t.time, * EXCEPT(time)
        FROM
          (
            SELECT
              TIMESTAMP_SECONDS(
                CAST(timestamp AS INT64)
              ) as time,
              avg(average_exec_per_sec) as average_exec_per_sec
            FROM `test-clusterfuzz`.parent_stats.TestcaseRun
            WHERE
              (
                _PARTITIONTIME BETWEEN TIMESTAMP_SECONDS(1475280000) AND
                TIMESTAMP_SECONDS(1475798400)
              ) AND
              (
                job = \'test_job\' OR job = \'test_job2\'
              ) AND
                fuzzer = \'parent_child\'
              GROUP BY time
            ) as t
        """))

  def test_query_invalid_names(self):
    """Tests passing invalid fuzzer/job names."""
    stats_columns = ('sum(j.testcases_executed) as testcases_executed, '
                     'sum(j.new_crashes) as new_crashes, '
                     'sum(j.known_crashes) as known_crashes, '
                     'avg(t.average_exec_per_sec) as average_exec_per_sec ')

    with self.assertRaises(fuzzer_stats.FuzzerStatsException):
      fuzzer_stats.TableQuery('fuzzer_n\'ame$', ['job_type'], stats_columns,
                              fuzzer_stats.QueryGroupBy.GROUP_BY_DAY,
                              datetime.date(2016, 10, 1),
                              datetime.date(2016, 10, 7))


@test_utils.with_cloud_emulators('datastore')
class BuiltinFieldTests(unittest.TestCase):
  """Builtin field tests."""

  def setUp(self):
    self.today = datetime.datetime.utcnow().date()
    self.yesterday = self.today - datetime.timedelta(days=1)
    cov_info = data_types.CoverageInformation(
        fuzzer='fuzzer1', date=self.yesterday)
    cov_info.edges_covered = 11
    cov_info.edges_total = 30
    cov_info.functions_covered = 10
    cov_info.functions_total = 15
    cov_info.html_report_url = 'https://report_for_fuzzer1/{}'.format(
        data_types.coverage_information_date_to_string(self.yesterday))
    cov_info.corpus_size_units = 20
    cov_info.corpus_size_bytes = 200
    cov_info.quarantine_size_units = 5
    cov_info.quarantine_size_bytes = 50
    cov_info.corpus_location = 'gs://corpus'
    cov_info.corpus_backup_location = 'gs://corpus-backup/file.zip'
    cov_info.quarantine_location = 'gs://quarantine'
    cov_info.put()

    cov_info = data_types.CoverageInformation(fuzzer='fuzzer2', date=self.today)
    cov_info.edges_covered = 16
    cov_info.edges_total = 33
    cov_info.functions_covered = 58
    cov_info.functions_total = 90
    cov_info.html_report_url = 'https://report_for_fuzzer2/{}'.format(
        data_types.coverage_information_date_to_string(self.today))
    cov_info.corpus_size_units = 40
    cov_info.corpus_size_bytes = 99
    cov_info.quarantine_size_units = 6
    cov_info.quarantine_size_bytes = 14
    cov_info.corpus_location = 'gs://corpus'
    cov_info.corpus_backup_location = 'gs://corpus-backup/file.zip'
    cov_info.quarantine_location = 'gs://quarantine'
    cov_info.put()

    cov_info = data_types.CoverageInformation(
        fuzzer='fuzzer2', date=self.yesterday)
    cov_info.edges_covered = 15
    cov_info.edges_total = 40
    cov_info.functions_covered = 11
    cov_info.functions_total = 16
    cov_info.html_report_url = 'https://report_for_fuzzer2/{}'.format(
        data_types.coverage_information_date_to_string(self.yesterday))
    cov_info.corpus_size_units = 15
    cov_info.corpus_size_bytes = 230
    cov_info.quarantine_size_units = 8
    cov_info.quarantine_size_bytes = 60
    cov_info.corpus_location = 'gs://corpus'
    cov_info.corpus_backup_location = 'gs://corpus-backup/file.zip'
    cov_info.quarantine_location = 'gs://quarantine'
    cov_info.put()

    cov_info = data_types.CoverageInformation(fuzzer='fuzzer3', date=self.today)
    cov_info.edges_covered = None
    cov_info.edges_total = None
    cov_info.functions_covered = None
    cov_info.functions_total = None
    cov_info.html_report_url = None
    cov_info.corpus_size_units = 0
    cov_info.corpus_size_bytes = 0
    cov_info.quarantine_size_units = 0
    cov_info.quarantine_size_bytes = 0
    cov_info.corpus_location = 'gs://corpus'
    cov_info.corpus_backup_location = 'gs://corpus-backup/file.zip'
    cov_info.quarantine_location = 'gs://quarantine'
    cov_info.put()

    data_types.Job(
        name='job1', environment_string='FUZZ_LOGS_BUCKET = bucket1').put()
    data_types.Job(
        name='job2', environment_string='FUZZ_LOGS_BUCKET = bucket2').put()

  def test_constructors(self):
    """Test builtin field constructors."""
    field = fuzzer_stats.BuiltinFieldSpecifier('_EDGE_COV').create()
    self.assertIsInstance(field, fuzzer_stats.CoverageField)

    field = fuzzer_stats.BuiltinFieldSpecifier('_FUNC_COV').create()
    self.assertIsInstance(field, fuzzer_stats.CoverageField)

    field = fuzzer_stats.BuiltinFieldSpecifier('_CORPUS_SIZE').create()
    self.assertIsInstance(field, fuzzer_stats.CorpusSizeField)

    field = fuzzer_stats.BuiltinFieldSpecifier('_CORPUS_BACKUP').create()
    self.assertIsInstance(field, fuzzer_stats.CorpusBackupField)

    field = fuzzer_stats.BuiltinFieldSpecifier('_QUARANTINE_SIZE').create()
    self.assertIsInstance(field, fuzzer_stats.CorpusSizeField)

    field = fuzzer_stats.BuiltinFieldSpecifier('_COV_REPORT').create()
    self.assertIsInstance(field, fuzzer_stats.CoverageReportField)

  def test_coverage_fields(self):
    """Test coverage fields."""
    ctx = fuzzer_stats.CoverageFieldContext()
    edge_field = fuzzer_stats.BuiltinFieldSpecifier('_EDGE_COV').create(ctx)
    func_field = fuzzer_stats.BuiltinFieldSpecifier('_FUNC_COV').create(ctx)

    data = edge_field.get(fuzzer_stats.QueryGroupBy.GROUP_BY_FUZZER, 'fuzzer1')
    self.assertEqual(data.value, '36.67% (11/30)')
    self.assertAlmostEqual(data.sort_key, 36.666666666666664)
    self.assertIsNone(data.link)

    data = func_field.get(fuzzer_stats.QueryGroupBy.GROUP_BY_FUZZER, 'fuzzer2')
    self.assertEqual(data.value, '64.44% (58/90)')
    self.assertAlmostEqual(data.sort_key, 64.44444444444444)
    self.assertIsNone(data.link)

    ctx = fuzzer_stats.CoverageFieldContext(fuzzer='fuzzer2')
    edge_field = fuzzer_stats.BuiltinFieldSpecifier('_EDGE_COV').create(ctx)
    data = edge_field.get(fuzzer_stats.QueryGroupBy.GROUP_BY_DAY, self.today)
    self.assertEqual(data.value, '48.48% (16/33)')
    self.assertAlmostEqual(data.sort_key, 48.484848484848484)
    self.assertIsNone(data.link)

    data = edge_field.get(fuzzer_stats.QueryGroupBy.GROUP_BY_DAY,
                          self.yesterday)
    self.assertEqual(data.value, '37.50% (15/40)')
    self.assertAlmostEqual(data.sort_key, 37.5)
    self.assertIsNone(data.link)

  def test_corpus_size_fields(self):
    """Test corpus size fields."""
    ctx = fuzzer_stats.CoverageFieldContext()
    corpus_field = fuzzer_stats.BuiltinFieldSpecifier('_CORPUS_SIZE').create(
        ctx)
    corpus_backup_field = fuzzer_stats.BuiltinFieldSpecifier(
        '_CORPUS_BACKUP').create(ctx)
    quarantine_field = fuzzer_stats.BuiltinFieldSpecifier(
        '_QUARANTINE_SIZE').create(ctx)

    data = corpus_field.get(fuzzer_stats.QueryGroupBy.GROUP_BY_FUZZER,
                            'fuzzer1')
    self.assertEqual(data.value, '20 (200 B)')
    self.assertEqual(data.sort_key, 20)
    self.assertEqual(data.link, 'gs://corpus')

    data = corpus_backup_field.get(fuzzer_stats.QueryGroupBy.GROUP_BY_FUZZER,
                                   'fuzzer1')
    self.assertEqual(data.value, 'Download')
    self.assertEqual(data.sort_key, None)
    self.assertEqual(data.link, 'gs://corpus-backup')

    data = corpus_field.get(fuzzer_stats.QueryGroupBy.GROUP_BY_FUZZER,
                            'fuzzer2')
    self.assertEqual(data.value, '40 (99 B)')
    self.assertEqual(data.sort_key, 40)
    self.assertEqual(data.link, 'gs://corpus')

    data = corpus_backup_field.get(fuzzer_stats.QueryGroupBy.GROUP_BY_FUZZER,
                                   'fuzzer2')
    self.assertEqual(data.value, 'Download')
    self.assertEqual(data.sort_key, None)
    self.assertEqual(data.link, 'gs://corpus-backup')

    data = quarantine_field.get(fuzzer_stats.QueryGroupBy.GROUP_BY_FUZZER,
                                'fuzzer1')
    self.assertEqual(data.value, '5 (50 B)')
    self.assertEqual(data.sort_key, 5)
    self.assertEqual(data.link, 'gs://quarantine')

    data = quarantine_field.get(fuzzer_stats.QueryGroupBy.GROUP_BY_FUZZER,
                                'fuzzer2')
    self.assertEqual(data.value, '6 (14 B)')
    self.assertEqual(data.sort_key, 6)
    self.assertEqual(data.link, 'gs://quarantine')

    ctx = fuzzer_stats.CoverageFieldContext('fuzzer2')
    corpus_field = fuzzer_stats.BuiltinFieldSpecifier('_CORPUS_SIZE').create(
        ctx)
    corpus_backup_field = fuzzer_stats.BuiltinFieldSpecifier(
        '_CORPUS_BACKUP').create(ctx)

    data = corpus_field.get(fuzzer_stats.QueryGroupBy.GROUP_BY_DAY, self.today)
    self.assertEqual(data.value, '40 (99 B)')
    self.assertEqual(data.sort_key, 40)
    self.assertEqual(data.link, 'gs://corpus')

    data = corpus_backup_field.get(fuzzer_stats.QueryGroupBy.GROUP_BY_DAY,
                                   self.today)
    self.assertEqual(data.value, 'Download')
    self.assertEqual(data.sort_key, None)
    self.assertEqual(data.link, 'gs://corpus-backup')

    data = corpus_field.get(fuzzer_stats.QueryGroupBy.GROUP_BY_DAY,
                            self.yesterday)
    self.assertEqual(data.value, '15 (230 B)')
    self.assertEqual(data.sort_key, 15)
    self.assertEqual(data.link, 'gs://corpus')

    data = corpus_backup_field.get(fuzzer_stats.QueryGroupBy.GROUP_BY_DAY,
                                   self.yesterday)
    self.assertEqual(data.value, 'Download')
    self.assertEqual(data.sort_key, None)
    self.assertEqual(data.link, 'gs://corpus-backup')

  def test_coverage_report_field(self):
    """Test coverage report field."""
    ctx = fuzzer_stats.CoverageFieldContext()
    coverage_report_field = fuzzer_stats.BuiltinFieldSpecifier(
        '_COV_REPORT').create(ctx)

    ctx = fuzzer_stats.CoverageFieldContext('fuzzer2')
    coverage_report_field = fuzzer_stats.BuiltinFieldSpecifier(
        '_COV_REPORT').create(ctx)

    data = coverage_report_field.get(fuzzer_stats.QueryGroupBy.GROUP_BY_DAY,
                                     self.today)
    self.assertEqual(data.value, 'Coverage')
    self.assertEqual(
        data.link, 'https://report_for_fuzzer2/{}'.format(
            data_types.coverage_information_date_to_string(self.today)))

    data = coverage_report_field.get(fuzzer_stats.QueryGroupBy.GROUP_BY_DAY,
                                     self.yesterday)
    self.assertEqual(data.value, 'Coverage')
    self.assertEqual(
        data.link, 'https://report_for_fuzzer2/{}'.format(
            data_types.coverage_information_date_to_string(self.yesterday)))

  def test_coverage_field_invalid_info(self):
    """Test that coverage field works as expected with invalid coverage info."""
    ctx = fuzzer_stats.CoverageFieldContext(fuzzer='fuzzer3')
    edge_field = fuzzer_stats.BuiltinFieldSpecifier('_EDGE_COV').create(ctx)
    data = edge_field.get(fuzzer_stats.QueryGroupBy.GROUP_BY_DAY, self.today)
    self.assertIsNone(data)

  def test_logs_field_by_fuzzer(self):
    """Test logs field (group by fuzzer)."""
    ctx = fuzzer_stats.FuzzerRunLogsContext('fuzzer1', ['job1'])
    logs_field = fuzzer_stats.BuiltinFieldSpecifier('_FUZZER_RUN_LOGS').create(
        ctx)

    data = logs_field.get(fuzzer_stats.QueryGroupBy.GROUP_BY_FUZZER,
                          'fuzzer_child1')
    self.assertEqual(data.value, 'Logs')
    self.assertEqual(data.link, 'gs://bucket1/fuzzer_child1/job1')

  def test_logs_field_by_day(self):
    """Test logs field (group by day)."""
    ctx = fuzzer_stats.FuzzerRunLogsContext('fuzzer1', ['job1'])
    logs_field = fuzzer_stats.BuiltinFieldSpecifier('_FUZZER_RUN_LOGS').create(
        ctx)

    data = logs_field.get(fuzzer_stats.QueryGroupBy.GROUP_BY_DAY,
                          datetime.date(2016, 11, 18))
    self.assertEqual(data.value, 'Logs')
    self.assertEqual(data.link, 'gs://bucket1/fuzzer1/job1/2016-11-18')

  def test_logs_field_by_job(self):
    """Test logs field (group by job)."""
    ctx = fuzzer_stats.FuzzerRunLogsContext('fuzzer1', ['blah'])
    logs_field = fuzzer_stats.BuiltinFieldSpecifier('_FUZZER_RUN_LOGS').create(
        ctx)

    data = logs_field.get(fuzzer_stats.QueryGroupBy.GROUP_BY_JOB, 'job2')
    self.assertEqual(data.value, 'Logs')
    self.assertEqual(data.link, 'gs://bucket2/fuzzer1/job2')

  def test_performance_field_by_fuzzer(self):
    """Test performance field (group by fuzzer)."""
    ctx = fuzzer_stats.FuzzerRunLogsContext('fuzzer1', ['job1'])
    performance_field = (
        fuzzer_stats.BuiltinFieldSpecifier('_PERFORMANCE_REPORT').create(ctx))

    data = performance_field.get(fuzzer_stats.QueryGroupBy.GROUP_BY_FUZZER,
                                 'fuzzer_child1')
    self.assertEqual(data.value, 'Performance')
    expected_link = '/performance-report/fuzzer_child1/job1/latest'
    self.assertEqual(data.link, expected_link)

  def test_performance_field_by_day(self):
    """Test performance field (group by day)."""
    ctx = fuzzer_stats.FuzzerRunLogsContext('fuzzer1', ['job1'])
    performance_field = (
        fuzzer_stats.BuiltinFieldSpecifier('_PERFORMANCE_REPORT').create(ctx))

    data = performance_field.get(fuzzer_stats.QueryGroupBy.GROUP_BY_DAY,
                                 datetime.date(2016, 11, 18))
    self.assertEqual(data.value, 'Performance')
    expected_link = '/performance-report/fuzzer1/job1/2016-11-18'
    self.assertEqual(data.link, expected_link)

  def test_performance_field_by_job(self):
    """Test performance field (group by job)."""
    ctx = fuzzer_stats.FuzzerRunLogsContext('fuzzer1', ['blah'])
    performance_field = (
        fuzzer_stats.BuiltinFieldSpecifier('_PERFORMANCE_REPORT').create(ctx))

    data = performance_field.get(fuzzer_stats.QueryGroupBy.GROUP_BY_JOB, 'job2')
    self.assertEqual(data.value, 'Performance')
    expected_link = '/performance-report/fuzzer1/job2/latest'
    self.assertEqual(data.link, expected_link)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Custom init runner."""

import os

from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.system import process_handler

SCRIPT_DIR = os.path.join('bot', 'init')


def _extension(platform):
  """Get the init extension for a platform."""
  if platform == 'windows':
    return '.ps1'

  return '.bash'


def run():
  """Run custom platform specific init scripts."""
  platform = environment.platform().lower()
  script_path = os.path.join(environment.get_config_directory(), SCRIPT_DIR,
                             platform + _extension(platform))
  if not os.path.exists(script_path):
    return

  os.chmod(script_path, 0o750)
  if script_path.endswith('.ps1'):
    cmd = 'powershell.exe ' + script_path
  else:
    cmd = script_path

  try:
    process_handler.run_process(
        cmd,
        timeout=1800,
        need_shell=True,
        testcase_run=False,
        ignore_children=True)
  except Exception:
    logs.log_error('Failed to execute platform initialization script.')



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for the Delta minimizer."""

from clusterfuzz._internal.bot.minimizer import delta_minimizer

from . import base_minimizer_tester


class DeltaMinimizerTest(base_minimizer_tester.BaseMinimizerTester):
  """Test for Delta Minimizer.Sets up the minimizer being used and then
  runs all of the tests in BaseMinimizerTester."""

  def setUp(self):
    self.line_minimizer = delta_minimizer.DeltaMinimizer(
        self._mock_test_function)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Run clean indexes on both internal and external."""

import os
import sys

from local.butler import common
from src.clusterfuzz._internal.config import local_config

INDEX_FILE_PATH = 'src/appengine/index.yaml'


def _cleanup_indexes(project, index_yaml_path):
  """Cleanup indexes."""
  common.execute(('gcloud datastore indexes cleanup '
                  '--quiet --project {project} {index_yaml_path}').format(
                      project=project, index_yaml_path=index_yaml_path))


def execute(args):
  """Clean indexes."""
  if not os.path.exists(args.config_dir):
    print('Please provide a valid configuration directory.')
    sys.exit(1)
  os.environ['CONFIG_DIR_OVERRIDE'] = args.config_dir

  config = local_config.GAEConfig()
  application_id = config.get('application_id')

  _cleanup_indexes(application_id, INDEX_FILE_PATH)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for environment."""

import os
import re
import unittest

import mock

from clusterfuzz._internal.bot.untrusted_runner import environment
from clusterfuzz._internal.tests.test_libs import helpers

FORWARDED_ENVIRONMENT_VARIABLES = [
    re.compile(pattern) for pattern in (
        r'^ASAN_OPTIONS$',
        r'^AFL_.*',
        r'^REBASED$',
    )
]

REBASED_ENVIRONMENT_VARIABLES = set([
    'FUZZER_DIR',
    'REBASED',
])


@mock.patch('clusterfuzz._internal.bot.untrusted_runner.environment.'
            'FORWARDED_ENVIRONMENT_VARIABLES', FORWARDED_ENVIRONMENT_VARIABLES)
@mock.patch(
    'clusterfuzz._internal.bot.untrusted_runner.environment.REBASED_ENVIRONMENT_VARIABLES',
    REBASED_ENVIRONMENT_VARIABLES)
class EnvironmentTest(unittest.TestCase):
  """Test environment."""

  def setUp(self):
    helpers.patch(self, [
        'clusterfuzz._internal.bot.untrusted_runner.host.stub',
    ])

    helpers.patch_environ(self)
    os.environ['WORKER_ROOT_DIR'] = '/worker'

  def test_is_forwarded_environment_variable(self):
    """Test is_forwarded_environment_variable."""
    self.assertTrue(
        environment.is_forwarded_environment_variable('ASAN_OPTIONS'))
    self.assertTrue(environment.is_forwarded_environment_variable('REBASED'))
    self.assertFalse(
        environment.is_forwarded_environment_variable('FUZZER_DIR'))
    self.assertFalse(
        environment.is_forwarded_environment_variable('ASAN_OPTIONSS'))
    self.assertTrue(environment.is_forwarded_environment_variable('AFL_'))
    self.assertTrue(environment.is_forwarded_environment_variable('AFL_BLAH'))

  def test_should_rebase_environment_value(self):
    """Test should_rebase_environment_value."""
    self.assertTrue(environment.should_rebase_environment_value('FUZZER_DIR'))
    self.assertTrue(environment.should_rebase_environment_value('REBASED'))
    self.assertFalse(
        environment.should_rebase_environment_value('ASAN_OPTIONS'))
    self.assertFalse(environment.should_rebase_environment_value('AFL_'))

  def test_update_environment(self):
    """Test update environment."""
    environment.update_environment({
        'BLAH': 'abc',
        'BLAH2': os.path.join(os.environ['ROOT_DIR'], 'blah2'),
        'FUZZER_DIR': os.path.join(os.environ['ROOT_DIR'], 'fuzzer'),
    })

    request = self.mock.stub().UpdateEnvironment.call_args[0][0]
    self.assertEqual({
        'BLAH': 'abc',
        'BLAH2': os.path.join(os.environ['ROOT_DIR'], 'blah2'),
        'FUZZER_DIR': '/worker/fuzzer',
    }, request.env)

  def test_set_environment_vars(self):
    """Test set_environment_vars."""
    result = {}
    environment.set_environment_vars(
        result, {
            'ASAN_OPTIONS': 'options',
            'BLAH': 'blah',
            'FUZZER_DIR': os.path.join(os.environ['ROOT_DIR'], 'fuzzer'),
            'REBASED': os.path.join(os.environ['ROOT_DIR'], 'rebased')
        })

    self.assertDictEqual({
        'ASAN_OPTIONS': 'options',
        'REBASED': os.path.join(os.environ['ROOT_DIR'], 'rebased')
    }, result)

    os.environ['TRUSTED_HOST'] = 'True'

    environment.set_environment_vars(
        result, {
            'ASAN_OPTIONS': 'options',
            'BLAH': 'blah',
            'FUZZER_DIR': os.path.join(os.environ['ROOT_DIR'], 'fuzzer'),
            'REBASED': os.path.join(os.environ['ROOT_DIR'], 'rebased')
        })

    self.assertDictEqual({
        'ASAN_OPTIONS': 'options',
        'REBASED': '/worker/rebased',
    }, result)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""App Engine helpers."""

from distutils import spawn
import os
import shutil
import sys

from local.butler import common
from local.butler import constants

SRC_DIR_PY = os.path.join('src', 'appengine')


def _add_env_vars_if_needed(yaml_path, additional_env_vars):
  """Add environment variables to yaml file if necessary."""
  # Defer imports since our python paths have to be set up first.
  import yaml

  from src.clusterfuzz._internal.config import local_config

  env_values = local_config.ProjectConfig().get('env')
  if additional_env_vars:
    env_values.update(additional_env_vars)

  if not env_values:
    return

  with open(yaml_path) as f:
    data = yaml.safe_load(f)

  if not isinstance(data, dict) or 'service' not in data:
    # Not a service.
    return

  data.setdefault('env_variables', {}).update(env_values)
  with open(yaml_path, 'w') as f:
    yaml.safe_dump(data, f)


def copy_yamls_and_preprocess(paths, additional_env_vars=None):
  """Copy paths to appengine source directories since they reference sources
  and otherwise, deployment fails."""
  rebased_paths = []
  for path in paths:
    target_filename = os.path.basename(path)
    rebased_path = os.path.join(SRC_DIR_PY, target_filename)

    # Remove target in case it's a symlink, since shutil.copy follows symlinks.
    if os.path.exists(rebased_path):
      os.remove(rebased_path)
    shutil.copy(path, rebased_path)
    os.chmod(rebased_path, 0o600)

    _add_env_vars_if_needed(rebased_path, additional_env_vars)
    rebased_paths.append(rebased_path)

  return rebased_paths


def find_sdk_path():
  """Find the App Engine SDK path."""
  if common.get_platform() == 'windows':
    _, gcloud_path = common.execute('where gcloud.cmd', print_output=False)
  else:
    gcloud_path = spawn.find_executable('gcloud')

  if not gcloud_path:
    print('Please install the Google Cloud SDK and set up PATH to point to it.')
    sys.exit(1)

  cloud_sdk_path = os.path.dirname(
      os.path.dirname(os.path.realpath(gcloud_path)))
  appengine_sdk_path = os.path.join(cloud_sdk_path, 'platform',
                                    'google_appengine')
  if not os.path.exists(appengine_sdk_path):
    print('App Engine SDK not found. Please run local/install_deps.bash')
    sys.exit(1)

  return appengine_sdk_path


def symlink_dirs():
  """Symlink folders for use on appengine."""
  symlink_config_dir()

  common.symlink(
      src=os.path.join('src', 'clusterfuzz'),
      target=os.path.join(SRC_DIR_PY, 'clusterfuzz'))

  # Remove existing local_gcs symlink (if any). This is important, as otherwise
  # we will try deploying the directory in production. This is only needed for
  # local development in run_server.
  local_gcs_symlink_path = os.path.join(SRC_DIR_PY, 'local_gcs')
  common.remove_symlink(local_gcs_symlink_path)


def build_templates():
  """Build template files used in appengine."""
  common.execute('python polymer_bundler.py', cwd='local')


def symlink_config_dir():
  """Symlink config directory in appengine directory."""
  config_dir = os.getenv('CONFIG_DIR_OVERRIDE', constants.TEST_CONFIG_DIR)
  common.symlink(src=config_dir, target=os.path.join(SRC_DIR_PY, 'config'))


def region_from_location(location):
  """Convert an app engine location ID to a region."""
  if not location[-1].isdigit():
    # e.g. us-central -> us-central1
    location += '1'

  return location


def region(project):
  """Get the App Engine region."""
  return_code, location = common.execute(
      'gcloud app describe --project={project} '
      '--format="value(locationId)"'.format(project=project))
  if return_code:
    raise RuntimeError('Could not get App Engine region')

  return region_from_location(location.strip().decode('utf-8'))



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A cron handler that batches FuzzerJobs."""

from google.cloud import ndb

from clusterfuzz._internal.datastore import data_types
from handlers import base_handler
from libs import handler

FUZZER_JOB_BATCH_SIZE = 4000


def batch_fuzzer_jobs():
  """Batch FuzzerJobs for reduced Datastore read ops by bots."""
  platforms = [
      item.platform for item in data_types.FuzzerJob.query(
          projection=[data_types.FuzzerJob.platform], distinct=True)
  ]

  for platform in platforms:
    fuzzer_jobs = list(
        data_types.FuzzerJob.query(data_types.FuzzerJob.platform == platform))
    fuzzer_jobs.sort(key=lambda item: item.job)

    batches_to_remove = set(
        b.key for b in data_types.FuzzerJobs.query(
            data_types.FuzzerJobs.platform == platform))

    batch_count = 0
    for i in range(0, len(fuzzer_jobs), FUZZER_JOB_BATCH_SIZE):
      key_id = platform + '-' + str(batch_count)
      end = min(i + FUZZER_JOB_BATCH_SIZE, len(fuzzer_jobs))

      batched = data_types.FuzzerJobs(id=key_id, platform=platform)
      batched.platform = platform
      batched.fuzzer_jobs = fuzzer_jobs[i:end]
      batched.put()
      batch_count += 1

      batches_to_remove.discard(batched.key)

    # Remove additional batches if number reduced.
    if batches_to_remove:
      ndb.delete_multi(batches_to_remove)


class Handler(base_handler.Handler):
  """Handler for building data_types.CrashsStats2."""

  @handler.cron()
  def get(self):
    """Process a GET request from a cronjob."""
    batch_fuzzer_jobs()



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Schedule corpus pruning tasks."""

from clusterfuzz._internal.base import tasks
from clusterfuzz._internal.base import utils
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.datastore import fuzz_target_utils
from handlers import base_handler
from libs import handler


def get_tasks_to_schedule():
  """Return (task_target, job_name, queue_name) arguments to schedule a task."""
  for job in data_types.Job.query():
    if not utils.string_is_true(job.get_environment().get('CORPUS_PRUNE')):
      continue

    queue_name = tasks.queue_for_job(job.name)
    for target_job in fuzz_target_utils.get_fuzz_target_jobs(job=job.name):
      task_target = target_job.fuzz_target_name
      yield (task_target, job.name, queue_name)


class Handler(base_handler.Handler):
  """Schedule corpus pruning tasks.."""

  @handler.cron()
  def get(self):
    """Schedule the corpus pruning tasks."""
    for task_target, job_name, queue_name in get_tasks_to_schedule():
      tasks.add_task('corpus_pruning', task_target, job_name, queue=queue_name)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""testcase_list tests."""
import datetime
import unittest

import flask
import mock
import webtest

from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.tests.test_libs import helpers as test_helpers
from clusterfuzz._internal.tests.test_libs import test_utils
from handlers import testcase_list
from libs import helpers
from libs.query import datastore_query


def _has_access(need_privileged_access=False):
  return not need_privileged_access


def _make_permissions(is_prefix, name):
  perm = data_types.ExternalUserPermission()
  perm.entity_name = name
  perm.is_prefix = is_prefix
  return perm


class AddFiltersTest(unittest.TestCase):
  """Test add_filters(..)."""

  def setUp(self):
    self.params = {}

    self.query = mock.Mock()

  def test_no_field(self):
    """Test no field."""
    testcase_list.add_filters(self.query, self.params)
    self.query.order.assert_called_once_with('timestamp', is_desc=True)
    self.query.filter.assert_has_calls([
        mock.call('status', 'Processed'),
        mock.call('is_a_duplicate_flag', False),
        mock.call('open', True),
        mock.call('is_leader', True)
    ])

  def test_both_fields(self):
    """Test filter field and keyword field."""
    self.params['q'] = ('hello group:456 issue:123 platform:windows stable:s.1'
                        ' beta:b.2 extended_stable:es.1 fuzzer:2 job:3')
    self.params['fuzzer'] = 'fuzz'
    self.params['issue'] = 'yes'
    self.params['job'] = 'somejob'
    self.params['open'] = 'yes'
    self.params['project'] = 'project'
    self.params['reproducible'] = 'yes'
    self.params['security'] = 'yes'
    self.params['impact'] = 'stable'
    testcase_list.add_filters(self.query, self.params)
    self.query.order.assert_called_once_with('timestamp', is_desc=True)
    self.query.filter.assert_has_calls([
        mock.call('status', 'Processed'),
        mock.call('is_a_duplicate_flag', False),
        mock.call('impact_version_indices', 'stable'),
        mock.call('has_bug_flag', True),
        mock.call('open', True),
        mock.call('security_flag', True),
        mock.call('group_id', 456),
        mock.call('bug_indices', '123'),
        mock.call('platform', 'windows'),
        mock.call('impact_extended_stable_version_indices', 'es.1'),
        mock.call('impact_stable_version_indices', 's.1'),
        mock.call('impact_beta_version_indices', 'b.2'),
        mock.call('fuzzer_name_indices', '2'),
        mock.call('job_type', '3'),
        mock.call('keywords', 'hello'),
        mock.call('one_time_crasher_flag', False),
        mock.call('job_type', 'somejob'),
        mock.call('fuzzer_name_indices', 'fuzz'),
        mock.call('project_name', 'project'),
    ])

  def test_revision_greater_than(self):
    """Ensure that we change ordering when using revision_greater_than."""
    self.params['revision_greater_than'] = '123456'
    testcase_list.add_filters(self.query, self.params)
    self.query.order.assert_called_once_with('crash_revision', is_desc=True)
    self.query.filter.assert_has_calls([
        mock.call('status', 'Processed'),
        mock.call('is_a_duplicate_flag', False),
        mock.call('is_leader', True),
        mock.call('crash_revision', 123456, operator='>'),
    ])


class GroupFilterTest(unittest.TestCase):
  """Test GroupFilter."""

  def setUp(self):
    self.filter = testcase_list.GroupFilter()
    self.query = mock.create_autospec(datastore_query.Query)

  def test_has_group(self):
    """Test having group."""
    self.filter.add(self.query, {'group': '1234'})
    self.query.filter.assert_has_calls([mock.call('group_id', 1234)])

  def test_no_group(self):
    """Test no group."""
    self.filter.add(self.query, {'group': ''})
    self.query.filter.assert_has_calls([mock.call('is_leader', True)])

  def test_raise_exception(self):
    """Test raising exceptions."""
    with self.assertRaises(helpers.EarlyExitException) as cm:
      self.filter.add(self.query, {'group': '123x4'})

    self.query.filter.assert_has_calls([])
    self.assertEqual("'group' must be int.", str(cm.exception))


@test_utils.with_cloud_emulators('datastore')
class JsonHandlerTest(unittest.TestCase):
  """Test Handler."""

  def setUp(self):
    test_helpers.patch(self, [
        'clusterfuzz._internal.base.external_users._allowed_entities_for_user',
        'clusterfuzz._internal.base.external_users._get_permissions_query_for_user',
        'libs.access.get_user_job_type',
        'libs.access.has_access',
        'libs.helpers.get_user_email',
    ])
    flaskapp = flask.Flask('testflask')
    flaskapp.add_url_rule('/', view_func=testcase_list.JsonHandler.as_view('/'))
    self.app = webtest.TestApp(flaskapp)

    self.testcases = []
    for i in range(10):
      t = data_types.Testcase()
      t.crash_type = ''
      t.crash_state = ''
      t.status = 'Processed'
      t.security_flag = True
      t.is_a_duplicate_flag = False
      t.one_time_crasher_flag = False
      t.open = True
      t.is_leader = True
      t.timestamp = datetime.datetime.fromtimestamp(100 - i)
      t.put()
      self.testcases.append(t)

  def test_domain_with_job_and_fuzzer(self):
    """Test domain user with permitted jobis and permitted fuzzer."""

    self.mock.get_user_job_type.return_value = 'job2'
    self.mock.get_user_email.return_value = 'test@test.com'
    self.mock.has_access.side_effect = _has_access
    self.mock._allowed_entities_for_user.side_effect = [['job'], ['fuzzer']]  # pylint: disable=protected-access
    self.mock._get_permissions_query_for_user.side_effect = [  # pylint: disable=protected-access
        [_make_permissions(True, 'job_')],
        [_make_permissions(False, 'fuzzer')],
    ]

    self.testcases[5].job_type = 'job'
    self.testcases[5].security_flag = True
    self.testcases[5].one_time_crasher_flag = False
    self.testcases[5].put()
    self.testcases[6].fuzzer_name = 'fuzzer'
    self.testcases[6].security_flag = True
    self.testcases[6].one_time_crasher_flag = False
    self.testcases[6].put()
    self.testcases[7].job_type = 'job2'
    self.testcases[7].security_flag = False
    self.testcases[7].one_time_crasher_flag = False
    self.testcases[7].put()
    self.testcases[8].security_flag = False
    self.testcases[8].one_time_crasher_flag = False
    self.testcases[8].put()

    resp = self.app.post_json('/', {'reproducible': 'yes'})

    self.assertEqual(200, resp.status_int)
    self.assertListEqual([
        self.testcases[5].key.id(), self.testcases[6].key.id(),
        self.testcases[7].key.id(), self.testcases[8].key.id()
    ], [item['id'] for item in resp.json['items']])



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""oss_fuzz_apply_ccs tests."""
import datetime
import unittest

import flask
import six
import webtest

from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.tests.test_libs import helpers as test_helpers
from clusterfuzz._internal.tests.test_libs import test_utils
from handlers.cron import oss_fuzz_apply_ccs
from libs.issue_management import issue_tracker_policy
from libs.issue_management import monorail
from libs.issue_management.monorail.issue import Issue

OSS_FUZZ_POLICY = issue_tracker_policy.IssueTrackerPolicy({
    'deadline_policy_message':
        'This bug is subject to a 90 day disclosure deadline. '
        'If 90 days elapse\n'
        'without an upstream patch, then the bug report will automatically\n'
        'become visible to the public.',
    'labels': {
        'reported': 'Reported-%YYYY-MM-DD%',
        'restrict_view': 'Restrict-View-Commit',
    },
    'status': {
        'assigned': 'Assigned',
        'duplicate': 'Duplicate',
        'fixed': 'Fixed',
        'new': 'New',
        'verified': 'Verified',
        'wontfix': 'WontFix'
    }
})

DEADLINE_NOTE = (
    'This bug is subject to a 90 day disclosure deadline. If 90 days elapse\n'
    'without an upstream patch, then the bug report will automatically\n'
    'become visible to the public.')


class IssueTrackerManager(object):
  """Mock issue tracker manager."""

  def __init__(self, project_name):
    self.project_name = project_name
    self.last_issue = None
    self.modified_issues = {}

  def save(self, issue, *args, **kwargs):  # pylint: disable=unused-argument
    """Save a issue."""
    self.modified_issues[issue.id] = issue


def get_original_issue(self, issue_id):
  """Get original issue."""
  issue_id = int(issue_id)

  issue = Issue()
  issue.open = True
  issue.itm = self._itm  # pylint: disable=protected-access
  issue.id = issue_id

  if issue_id == 1337:
    issue.add_cc('user@example.com')
    issue.add_label('Restrict-View-Commit')
  elif issue_id == 1338:
    issue.add_cc('user@example.com')
    issue.add_cc('user2@example.com')
  elif issue_id == 1340:
    issue.add_label('reported-2015-01-01')

  return monorail.Issue(issue)


@test_utils.with_cloud_emulators('datastore')
class OssFuzzApplyCcsTest(unittest.TestCase):
  """Test OssFuzzApplyCcs."""

  def setUp(self):
    test_helpers.patch_environ(self)
    flaskapp = flask.Flask('testflask')
    flaskapp.add_url_rule(
        '/apply-ccs',
        view_func=oss_fuzz_apply_ccs.Handler.as_view('/apply-ccs'))
    self.app = webtest.TestApp(flaskapp)

    data_types.ExternalUserPermission(
        email='user@example.com',
        entity_name='job',
        entity_kind=data_types.PermissionEntityKind.JOB,
        is_prefix=False,
        auto_cc=data_types.AutoCCType.ALL).put()

    data_types.ExternalUserPermission(
        email='user2@example.com',
        entity_name='job',
        entity_kind=data_types.PermissionEntityKind.JOB,
        is_prefix=False,
        auto_cc=data_types.AutoCCType.ALL).put()

    test_helpers.patch(self, [
        'clusterfuzz._internal.base.utils.utcnow',
        'handlers.base_handler.Handler.is_cron',
        'libs.issue_management.issue_tracker.IssueTracker.get_original_issue',
        'libs.issue_management.issue_tracker_policy.get',
        'libs.issue_management.issue_tracker_utils.'
        'get_issue_tracker_for_testcase',
    ])

    self.itm = IssueTrackerManager('oss-fuzz')
    self.mock.get_issue_tracker_for_testcase.return_value = (
        monorail.IssueTracker(self.itm))
    self.mock.utcnow.return_value = datetime.datetime(2016, 1, 1)
    self.mock.get.return_value = OSS_FUZZ_POLICY
    self.mock.get_original_issue.side_effect = get_original_issue

    data_types.Testcase(
        open=True, status='Processed', bug_information='1337',
        job_type='job').put()

    data_types.Testcase(
        open=True, status='Processed', bug_information='1338',
        job_type='job').put()

    data_types.Testcase(
        open=True, status='Processed', bug_information='1339',
        job_type='job').put()

    data_types.Testcase(
        open=True, status='Processed', bug_information='1340',
        job_type='job').put()

  def test_execute(self):
    """Tests executing of cron job."""
    self.app.get('/apply-ccs')
    self.assertEqual(len(self.itm.modified_issues), 3)

    issue_1337 = self.itm.modified_issues[1337]
    six.assertCountEqual(self, issue_1337.cc, [
        'user@example.com',
        'user2@example.com',
    ])

    self.assertTrue(issue_1337.has_label_matching('reported-2016-01-01'))
    self.assertEqual(issue_1337.comment, DEADLINE_NOTE)

    self.assertNotIn(1338, self.itm.modified_issues)

    issue_1339 = self.itm.modified_issues[1339]
    six.assertCountEqual(self, issue_1339.cc, [
        'user@example.com',
        'user2@example.com',
    ])

    self.assertTrue(issue_1339.has_label_matching('reported-2016-01-01'))
    self.assertEqual(issue_1339.comment, '')

    issue_1340 = self.itm.modified_issues[1340]
    six.assertCountEqual(self, issue_1340.cc, [
        'user@example.com',
        'user2@example.com',
    ])
    self.assertTrue(issue_1340.has_label_matching('reported-2015-01-01'))
    self.assertFalse(issue_1340.has_label_matching('reported-2016-01-01'))
    self.assertEqual(issue_1340.comment, '')



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""find_similar_issues tests."""
import unittest

import flask
import mock
import webtest

from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.tests.test_libs import helpers as test_helpers
from clusterfuzz._internal.tests.test_libs import test_utils
from handlers.testcase_detail import find_similar_issues
from libs.issue_management import monorail
from libs.issue_management.monorail import issue


@test_utils.with_cloud_emulators('datastore')
class HandlerTest(unittest.TestCase):
  """Test FindSimilarIssuesHandler."""

  def setUp(self):
    test_helpers.patch(self, [
        'libs.access.check_access_and_get_testcase',
        'libs.issue_management.issue_tracker_utils.'
        'get_issue_tracker_for_testcase',
        'libs.issue_management.issue_tracker_utils.get_issue_url',
        'libs.issue_management.issue_tracker_utils.get_search_keywords',
        'libs.issue_management.issue_tracker_utils.get_similar_issues',
        'libs.issue_management.issue_tracker_utils.get_similar_issues_url',
    ])
    flaskapp = flask.Flask('testflask')
    flaskapp.add_url_rule(
        '/', view_func=find_similar_issues.Handler.as_view('/'))
    self.app = webtest.TestApp(flaskapp)

    self.testcase = data_types.Testcase()
    self.testcase.put()
    self.mock.check_access_and_get_testcase.return_value = self.testcase

    self.invalid_testcase_id = self.testcase.key.id() + 1

  def test_itm_not_found(self):
    """Ensure it errors when issue_tracker_manager doesn't exist."""
    self.mock.get_issue_tracker_for_testcase.return_value = None

    response = self.app.get(
        '/?testcaseId=%d&filterType=open' % self.testcase.key.id(),
        expect_errors=True)
    self.assertEqual(response.status_int, 404)
    self.mock.get_issue_tracker_for_testcase.assert_has_calls(
        [mock.call(mock.ANY)])
    self.assertEqual(
        self.testcase.key.id(),
        self.mock.get_issue_tracker_for_testcase.call_args[0][0].key.id())

  def test_find(self):
    """Ensure it returns correct JSON when everything is ok."""
    issue_tracker = mock.Mock()
    monorail_issue = issue.Issue()
    monorail_issue.id = 100
    issue_item = monorail.Issue(monorail_issue)
    self.mock.get_issue_tracker_for_testcase.return_value = issue_tracker
    self.mock.get_search_keywords.return_value = ['query']
    self.mock.get_similar_issues_url.return_value = 'similarurl'
    self.mock.get_similar_issues.return_value = [issue_item]
    issue_tracker.issue_url.return_value = 'issueurl'

    response = self.app.get(
        '/?testcaseId=%d&filterType=open' % self.testcase.key.id())
    self.assertEqual(response.status_int, 200)
    self.assertEqual(response.json['queryString'], 'query')
    self.assertEqual(response.json['queryUrl'], 'similarurl')
    self.assertEqual(len(response.json['items']), 1)

    self.assertEqual(response.json['items'][0]['issue']['id'], issue_item.id)
    issue_tracker.issue_url.assert_called_with(issue_item.id)

    self.mock.get_issue_tracker_for_testcase.assert_has_calls(
        [mock.call(mock.ANY)])
    self.assertEqual(
        self.testcase.key.id(),
        self.mock.get_issue_tracker_for_testcase.call_args[0][0].key.id())
    self.mock.get_similar_issues.assert_has_calls(
        [mock.call(issue_tracker, mock.ANY, only_open=True)])
    self.assertEqual(self.testcase.key.id(),
                     self.mock.get_similar_issues.call_args[0][1].key.id())
